"""game.py

Part of the Tundra project.
(C) 2006 Rebuild Software
http://rebui.ld.net.au

Contains absolutely nothing of interest. OR DOES IT?
"""

import globals as g
from excepts import *

from pygame.locals import *
import pygame 

import os, sys

import font, level, dialog, music, lemming, mouse, sprite, player

from pgu import gui


class Game:
    def __init__(self, collection, diffNum, levelNum, onQuit=None):
        self.staticBg = renderGui()#colour=(0,0,48,255))
        self.collection = -1
        self.onQuit = onQuit

        self.music = music.loadPlaylist(g.MUSIC_PLAYLIST)

        self.changeCurrentLevel(collection, diffNum, levelNum)
        self._makeGui()

        self.scrollLeft = Rect(0,g.TOPBAR_HEIGHT,20,g.PLAYAREA_HEIGHT)
        self.scrollRight = Rect(0,g.TOPBAR_HEIGHT,20,g.PLAYAREA_HEIGHT)
        self.scrollRight.right = g.SCREEN_WIDTH
    
        self.scrollFactor = 1

        self.repaintBg = True

        self.upperAnims = sprite.SpriteGroup()
        self.lowerAnims = sprite.SpriteGroup()

        self.protagonists = sprite.SpriteGroup()
    
        self.refresh()

        self.dialogIsActive = False
        self.paused = True
        self.pauseRendering = False

        # TODO: make the dialog optional?
        self.levelDialog.openSlightlyRaised()
        return

    def changeCurrentLevel(self, collection, diffNum, levelNum):
        if not collection in g.levelDB:
            raise ValueError, "Collection "+collection+" does not exist"
        elif not diffNum in g.levelDB[collection]:
            raise ValueError, "Difficulty level "+repr(diffNum)+" does not exist in collection "+collection
        elif not levelNum in g.levelDB[collection][diffNum]:
            raise ValueError, "Level "+repr(levelNum)+"+1 not found in difficulty "+repr(diffNum)+" of collection "+collection
        
        self.levelObj = level.loadLevel(collection, diffNum, levelNum)

        if not collection == self.collection:
            self.music.setCollection(collection)
        self.music.next()

        g.START_COLLECTION = collection # save last touched collection
        
        if self.levelObj.keyword: #special level!
            self.music.next(self.levelObj.keyword)
        
        self.collection = collection
        self.diffNum = diffNum
        self.levelNum = levelNum
#        self.levelNum = level.normalizeLevel(collection, diffNum, levelNum)
        g.skillCache.updateDirt(self.getFavColour())

        self.releaseRateMin = self.levelObj.releaseRate
        self.releaseRateCurr = self.levelObj.releaseRate

        return
    
    def refresh(self):

        testRect = Rect(self.levelObj.startPosX, 0, g.SCREEN_WIDTH_DEFAULT/2, g.PLAYAREA_HEIGHT/2)
        self.levelRect = Rect(0, 0, g.SCREEN_WIDTH/2, g.PLAYAREA_HEIGHT/2)
        if (testRect.centerx - self.levelRect.w/2 <0):
            self.levelRect.left = 0
        elif (testRect.centerx + self.levelRect.w/2 >= g.MAX_LEVEL_WIDTH):
            self.levelRect.right = g.MAX_LEVEL_WIDTH-1
        else:
            self.levelRect.center = testRect.center
        self.releaseRateMin = self.levelObj.releaseRate
        self.releaseRateCurr = self.levelObj.releaseRate
        self.timer = self.levelObj.timeLimit
        self.reverseTimer = 0
        self.tickBuffer = 0

        self.numLemmingsReleased = 0
        self.numLemmingsSaved = 0
        self.numLemmingsTotal = 0

        self.hatchesOpen = False
        self.hatchesActive = False

        self.fastForward = 1 # do one update per screen refresh
        
        self.warDeclared = False # nuke mode
        self.warIndex = 0
        self.warList = None

        if hasattr(self, "skillbar"):
            self._refreshSkillbar()
        self.updateNotebar()

        self.foreground = self.levelObj.printTerrain()
        self.minimap.setMap(self.foreground)
        self.minimap.value = self.levelRect.left
        self.repaintBg = True

        if not hasattr(self, "preScaling"):
            self.preScaling = pygame.Surface(self.levelRect.size,SWSURFACE|SRCALPHA).convert_alpha()
            self.preScaling.fill((0,0,0,0))
        else:
            self.preScaling.fill((0,0,0,0))
        self.upperAnims.empty()
        self.lowerAnims.empty()

        self.protagonists.empty()

        self.spawnPoints = []
        self.trapRects = []
        self.spawnTimer = 0
        self.currentSpawnPoint = 0

        for i in self.levelObj.interacts:
# TODO: animation behavior is to loop
            if i.objType == "entrance":
                a = sprite.EntranceSprite(i, self)
                x,y = a.rect.midbottom
                self.spawnPoints.append((x,y-12))
            elif i.objType == "trap":
                a = sprite.TrapSprite(i, self)
                tempRect = Rect(i.activeRect)
                tempRect.x += i.posX
                tempRect.y += i.posY                
                self.trapRects.append((tempRect, a))
            else: #if all else fails, loop the animation
                a = sprite.HazardSprite(i, self)
            if not i.drawButDontOverwrite:
                self.upperAnims.add(a)
            else:
                self.lowerAnims.add(a)
        
        return

    def subtract(self, source, destpos):
        def checkPos((x,y),attr):
            try:
                result = self.collisionMask.get_at((x, y))
                value = g.MAP_KEY.get(attr,(0,0))
                return result[value[0]]&value[1] == value[1]
            except IndexError:
                raise OutOfThePlayfield # run away! run away!
        
        # must contain 2 colours: transparent and g.DEFAULT_COLOURKEY
        # delete terrain
        # modify collision mask
        r = source.get_rect()
        r.left = destpos[0]
        r.top = destpos[1]
        self.collisionMask.lock()
        self.foreground.lock()
        source.lock()

        for i in range(r.top, r.bottom):
            for j in range(r.left, r.right):
                if source.get_at((j-destpos[0],i-destpos[1])) != (0,0,0,0):
                    try:
                        p = ow,ox,oy,oz = self.collisionMask.get_at((j,i))                
                        if checkPos((j,i), "steel"):
                            continue

                        # TODO: terrible hack, should search for all drawonterrain objects
                        # right now I don't give a shit as I have no idea if pygame supports
                        # alpha masking or not and I don't want to dig my hole much deeper
                        if checkPos((j,i), "leftArrows") or checkPos((j,i), "rightArrows"):
                            for a in self.upperAnims:
                                if not hasattr(a,"inter"): continue
                                elif (a.inter.objType == "leftArrows" or a.inter.objType == "rightArrows") and a.rect.collidepoint(j,i):
                                    a.mask.fill( g.DEFAULT_COLOURKEY , ((j-a.rect.left,i-a.rect.top),(1,1)))
                                    a.unmaskedFrames = a.frameCount 
                                     
                        ow = ow & ~128 # remove terrain
                        self.foreground.fill( (0,0,0,0) , ((j,i),(1,1)) )
                        self.collisionMask.fill( (ow,ox,oy,oz) , ((j,i),(1,1)) )
                    except IndexError:
                        continue                    

        source.unlock()
        self.foreground.unlock()
        self.collisionMask.unlock()

    def update(self):
        for i in xrange(self.fastForward):
            if not self.paused:
                self.tickBuffer += 1
                if (self.timer == 0):
                    self.endGame(g.msg("finishDialog_timesup"))
                    return
                elif (self.numLemmingsReleased == 0) and (self.warDeclared or (self.numLemmingsTotal == self.levelObj.numReleased)):
                    self.endGame()
                    return                
    
                if (self.tickBuffer/g.GAMESECOND) > 1:
                    self.timer -= 1
                    self.reverseTimer += 1
                    if self.reverseTimer == 2:
                        self.hatchesOpen = True
                    elif self.reverseTimer == 3:
                        self.hatchesActive = True
                        self.music.play()
                    self.tickBuffer = 0
                
                if self.hatchesActive and not (len(self.spawnPoints) == 0) and (self.numLemmingsTotal < self.levelObj.numReleased): 
                # lemming insertion fork
                    self._insertLemming()
                
                if self.warDeclared and self.warIndex < len(self.warList):
                    try:
                        self.warList[self.warIndex].setSkill("s_bomber", ignoreDeathAction=True)
                    except SkillUnsettable:
                        pass
                    self.warIndex += 1
                    
                if not self.pauseRendering and not self.dialogIsActive:
                    self.protagonists.update()
            
            if not self.pauseRendering and (not self.paused or self.dialogIsActive):
                self.lowerAnims.update()
                self.upperAnims.update()
                

        if not self.dialogIsActive: # removing this line breaks the dialog scrollbars for some reason
            self.updateNotebar()

        return
                
    def draw(self, screen):        
        #stingeFilter = lambda rects, clippy: [r for r in rects if r.colliderect(clippy)]
     #   preScaleToScreen = lambda rects: [Rect(2*r.x, 2*r.y, 2*r.w, 2*r.h).move(0,g.TOPBAR_HEIGHT) for r in rects]
        
        rectList = []
        
        screen.blit(self.staticBg, (0,0))
        if not self.pauseRendering:
#            if not self.paused or self.dialogIsActive:
#                self.lowerAnims.update()
#                self.upperAnims.update()

#            if not self.paused and not self.dialogIsActive:
#                self.protagonists.update()
            
            #HACK: Force non-conservative updates on
#            self.repaintBg = True    

#            if self.repaintBg:
            self.preScaling.fill((0,0,0,0))
#            else:
#                rectList += self.lowerAnims.dirtyRects(self.preScaling.get_rect(), self.levelRect)
#                rectList += self.protagonists.dirtyRects(self.preScaling.get_rect(), self.levelRect)
#                rectList += self.upperAnims.dirtyRects(self.preScaling.get_rect(), self.levelRect)
#                rectList = stingeFilter(rectList, self.preScaling.get_rect())
#                for r in rectList:
#                    self.preScaling.fill((0,0,0,0), r)

            self.lowerAnims.drawShift(self.preScaling, self.levelRect)
            
#            if self.repaintBg:
#                self.repaintBg = False
            self.preScaling.blit(self.foreground, (0,0), self.levelRect)
#            else:
#                for r in rectList:
#                    self.preScaling.blit(self.foreground, r, r.move(self.levelRect.x, self.levelRect.y))
            # blit bg for a in spriteList, blit background segments
            self.upperAnims.drawShift(self.preScaling, self.levelRect)
            self.protagonists.drawShift(self.preScaling, self.levelRect)
            

        if g.ENABLE_SMOOTHING:
            screen.blit(pygame.transform.scale2x(self.preScaling), (0,g.TOPBAR_HEIGHT))
        else:
            screen.blit(pygame.transform.scale(self.preScaling, (g.SCREEN_WIDTH, g.PLAYAREA_HEIGHT)), (0,g.TOPBAR_HEIGHT))

        self.app.paint(screen)
        
     #   if self.dialogIsActive:
     #       return [g.SCREEN_RECT_PLAYAREA]
    
     #   return preScaleToScreen(rectList)+[g.SCREEN_RECT_PANEL] 

        
    def events(self):
        mbdown = False
        for e in pygame.event.get():
            if e.type is QUIT: player.saveLevelDB(g.CONFIG_FILE); raise SystemExit
#                if self.onQuit == None: raise SystemExit
#                else: self.onQuit(); return
            if e.type is KEYDOWN:
                if e.key == K_ESCAPE: 
                    if self.dialogIsActive: 
                        if self.onQuit == None: raise SystemExit
                        else: self.onQuit(); return
                    else: self.endGame(g.msg("finishDialog_giveup"))
                elif e.key == K_p:
                    self.paused = not self.paused
                elif e.key == K_f:
                    self.faster()
            if e.type is MOUSEBUTTONDOWN :
                if e.button == 3: self.scrollFactor = 2
                elif e.button == 1: mbdown = True
            if e.type is MOUSEBUTTONUP and e.button == 3: self.scrollFactor = 1 
            self.app.event(e)
        
        m = pygame.mouse.get_pos()
        if not (self.levelRect.left == 0) and self.scrollLeft.collidepoint(m):
            self.repaintBg = True
            if not (self.levelRect.left/(g.SCROLL_SPEED*self.scrollFactor)) == 0:
                self.levelRect.left -= g.SCROLL_SPEED*self.scrollFactor
            else:
                self.levelRect.left = 0
            self.minimap.value = self.levelRect.left
        elif not (self.levelRect.right == g.MAX_LEVEL_WIDTH) and self.scrollRight.collidepoint(m):
            self.repaintBg = True
            if not (g.MAX_LEVEL_WIDTH - self.levelRect.right)/(g.SCROLL_SPEED*self.scrollFactor) == 0:
                self.levelRect.right += (g.SCROLL_SPEED*self.scrollFactor)
            else:
                self.levelRect.right = g.MAX_LEVEL_WIDTH
            self.minimap.value = self.levelRect.left

        if not self.dialogIsActive:
            if self.skillbar.plus.state == 1:
                self.adjustReleaseRate(g.RELEASERATE_ADJUST)
            elif self.skillbar.minus.state == 1:
                self.adjustReleaseRate(-g.RELEASERATE_ADJUST)            

            lemCounter = 0
            target = None
            p = (m[0]/2+self.levelRect.left, (m[1]-g.TOPBAR_HEIGHT)/2)
            for l in self.protagonists:
                if l.rect.collidepoint(p) and not l.currentSkill == "s_ghost": # the lemming is not there
                    lemCounter += 1
                    if (not target) or (g.NOTEBAR_ORDER[target.getDisplayName()] <= g.NOTEBAR_ORDER[l.getDisplayName()]):
                        target = l
            
            if lemCounter > 0:
                if mbdown and (self.skillbarMap.get(self.skillGroup.value, (0,0))[0] > 0): #select-a-lemming
                    try:
                        if target.currentSkill == self.skillGroup.value:
                            raise SkillUnsettable                        
                        target.setSkill(self.skillGroup.value)
                        self.decrementSkill(self.skillGroup.value)
                    except SkillUnsettable:
                        pass
                self.notebar.hoverText = g.msg(target.getDisplayName())+" "+repr(lemCounter)
                g.mouse.changeCursor("select")
            else:
                self.notebar.hoverText = "" 
                g.mouse.changeCursor("pointer")            
        return

    def pause(self, *args):
        self.paused = not self.paused
        return

    def nuke(self, *args):
        self.warDeclared = True
        self.hatchesActive = False 
        self.warList = self.protagonists.sprites()       
        return
    
    def faster(self):
        if self.fastForward > 1:
            self.fastForward = 1
        else:
            self.fastForward = g.FASTER_MULTIPLIER
        return

    def getFavColour(self):
        return g.styleCache.get(self.levelObj.defaultStyle).favColour

    def _insertLemming(self):
        if self.spawnTimer==0:
            self.protagonists.add(lemming.Lemming(self.spawnPoints[self.currentSpawnPoint],self))
            self.numLemmingsTotal += 1
            self.numLemmingsReleased += 1

            self.currentSpawnPoint += 1
            self.currentSpawnPoint %= len(self.spawnPoints)

        self.spawnTimer += 1
        self.spawnTimer %= ((99-self.releaseRateCurr)/2 + 4)
        return

    # Used to preload time-consuming things not needed by the preview     
    def initializeLevelFinal(self):
        self.collisionMask = self.levelObj.printCollisionMask()
        return

    def endGame(self, message=g.msg("finishDialog_default")):
        self.updateNotebar()        
        self.paused = True
        self.dialogIsActive = True
        self.fastForward = 1
        self.music.fadeout()
        self.finishDialog.openSlightlyRaised(message)
        return

    def refreshMaskedSprites(self): 
        # imagine the look on an assembly programmer's face if he ever were to see this...
        for s in [a for a in self.upperAnims if a.isMasked] + [a for a in self.lowerAnims if a.isMasked]:
            i = s.inter            
            currStyle = g.styleCache.get(i.style)
            s.inter.frames = [pygame.transform.flip(x, False, i.drawUpsideDown) for x in currStyle.interDefs[i.itemID].frames]
        return
                


    def _makeTopbar(self):
        topbar = gui.Table(width=g.SCREEN_WIDTH,height=g.TOPBAR_HEIGHT)
        topbar.tr()
        topbar.td(gui.Label(""),width=g.SCREEN_WIDTH-3*40)
        
        p = font.Icon(pygame.image.load(g.RESOURCES_PATH+"pixmaps/topbar-pause.png") )
        p.connect(gui.CLICK, self.pause, None)
        topbar.td(p, width=40)

        f = font.Icon(pygame.image.load(g.RESOURCES_PATH+"pixmaps/topbar-fastfwd.png") )
        f.connect(gui.CLICK, self.faster)
        topbar.td(f, width=40)
        
        n = font.Icon(pygame.image.load(g.RESOURCES_PATH+"pixmaps/topbar-nuke.png") )
        n.connect(gui.CLICK, self.nuke, None)        
        topbar.td(n, width=40)
        return topbar

    def _refreshSkillbar(self):
        self.skillbar.clear()
        self.skillbar.tr()
        self.skillGroup.value = g.SKILL_ORDER[0]
        self.skillbar.minus = font.PixmapButton(gui.Image(font.skillButtonImage("s_minus", self.releaseRateMin)))
        self.skillbar.td(self.skillbar.minus, width=24)
        self.skillbar.plus = font.PixmapButton(gui.Image(font.skillButtonImage("s_plus", self.releaseRateCurr)))
        self.skillbar.td(self.skillbar.plus, width=24)

        for x in g.SKILL_ORDER:
            n = self.levelObj.numSkills.get(x,0)
            e = font.skillButtonImage(x, n)
            f = gui.Tool(self.skillGroup, gui.Image(e), x)
            #f.connect(gui.CLICK, self.decrementSkill, x)
            self.skillbarMap[x] = [n,f]
            self.skillbar.td(f, width=24)

        #e = gui.Button("TEST - Dialog")
#        e.connect(gui.CLICK,self.endGame,g.msg("finishDialog_default"))
        self.skillbar.td(self.minimap)
        self.skillbar.resize()
        return

    def _makeSkillbar(self):
        self.skillbar = gui.Table(width=g.SCREEN_WIDTH, height=g.SKILLBAR_HEIGHT)
        self.minimap = font.Minimap(self)
        self.skillGroup = gui.Group()
        self.skillbarMap = {}
        # TODO: link it up to the level object
        
        self._refreshSkillbar()

        return

    def decrementSkill(self, skillName):
        x = self.skillbarMap.get(skillName,None)
        if x and (x[0] > 0):
            x[0] -= 1
            x[1].setwidget(gui.Image(font.skillButtonImage(skillName, x[0])))
            x[1].resize()
        return

    def getScore(self):
#        if self.levelObj.numReleased == self.levelObj.numToSave:
#            return 5000+10*self.timer
#        else:
#            return 1000+int(4000*(self.numLemmingsSaved - self.levelObj.numToSave)/(self.levelObj.numReleased - self.levelObj.numToSave))+10*self.timer
        saveScore = self.numLemmingsSaved*6000/self.levelObj.numReleased
        skillScore = sum([y[0] for y in self.skillbarMap.itervalues()])*1200/sum(self.levelObj.numSkills.itervalues())
        timeScore = self.timer*600/self.levelObj.timeLimit
        #print saveScore, skillScore, timeScore
        return (saveScore+skillScore+timeScore)

    def updateNotebar(self, outAsFraction=True, inAsFraction=False):
        self.notebar.clear()
        self.notebar.tr()
        self.notebar.td(gui.Label(""),col=0,row=0, align=-1,width=240)
        self.notebar.td(gui.Image(font.stringToBitmap(self.notebar.hoverText)),col=0,row=0, align=-1,width=240)

        if outAsFraction:
            self.notebar.td(gui.Image(font.stringToBitmap("Out: "+repr(self.numLemmingsReleased)+"/"+repr(self.levelObj.numReleased))),align=-1,width=140)
        else:
            per = repr(int(100*float(self.numLemmingsReleased)/self.levelObj.numReleased))
            self.notebar.td(gui.Image(font.stringToBitmap("Out: "+per+"%")),align=-1,width=140)
        if inAsFraction:
            self.notebar.td(gui.Image(font.stringToBitmap("In: "+repr(self.numLemmingsSaved)+"/"+repr(self.levelObj.numReleased))),align=-1,width=140)
        else:
            per = repr(int(100*float(self.numLemmingsSaved)/self.levelObj.numReleased))
            self.notebar.td(gui.Image(font.stringToBitmap("In: "+per+"%")),align=-1,width=120)
        
        self.notebar.td(gui.Image(font.stringToBitmap("Time: "+repr(self.timer/60)+";"+repr(100+(self.timer%60))[-2:] )), align=-1, row=0, col=3, width=100)
        self.notebar.resize()
        return

    def adjustReleaseRate(self, increment):
        if increment > 0 and self.releaseRateCurr != g.RELEASERATE_MAX:
            if self.releaseRateCurr + increment > g.RELEASERATE_MAX:
                self.releaseRateCurr = g.RELEASERATE_MAX
            else:
                self.releaseRateCurr += increment
            self.skillbar.plus.setwidget(gui.Image(font.skillButtonImage("s_plus", self.releaseRateCurr)))
            # sound effect        
            
        elif increment < 0 and self.releaseRateCurr != self.releaseRateMin:
            if self.releaseRateCurr + increment < self.releaseRateMin:
                self.releaseRateCurr = self.releaseRateMin
            else:
                self.releaseRateCurr += increment
            self.skillbar.plus.setwidget(gui.Image(font.skillButtonImage("s_plus", self.releaseRateCurr)))
            # sound effect        
        return
            

    def _makeGui(self):
        self.app = gui.App()
        self.app.connect(gui.QUIT,self.app.quit,None)
    
        c = gui.Table(width=g.SCREEN_WIDTH,height=g.SCREEN_HEIGHT)
        self.topbar = self._makeTopbar()

        c.tr()
        c.td(self.topbar, valign=-1, height=g.TOPBAR_HEIGHT)
    
        self.levelDialog = dialog.LevelDialog(self)
        self.finishDialog = dialog.FinishDialog(self)
            
        c.tr()
        c.td(gui.Label(""), height = g.PLAYAREA_HEIGHT)
      
        self.notebar = gui.Table(width=g.SCREEN_WIDTH-32,height=g.NOTEBAR_HEIGHT)
        
        c.tr()
        c.td(self.notebar,height=g.NOTEBAR_HEIGHT)
        self.notebar.hoverText = ""

        self._makeSkillbar()

        c.tr()
        c.td(self.skillbar,height=g.SKILLBAR_HEIGHT)

        self.app.init(c)
        return

"""renderGui() -> Surface

Creates the noninteractive part of the Tundra interface which only needs to be rendered once.
"""
def renderGui(colour=(0,0,0,255)):
    surface = pygame.Surface([g.SCREEN_WIDTH, g.SCREEN_HEIGHT],SWSURFACE|SRCALPHA).convert_alpha()
    surface.fill(colour)
    topbar = pygame.image.load(g.RESOURCES_PATH + "pixmaps/topbar.png")
    skillbar = pygame.image.load(g.RESOURCES_PATH + "pixmaps/skillbar.png")
    bg = pygame.image.load(g.RESOURCES_PATH + "pixmaps/tempbg.png")    

    for i in range(0,surface.get_width(),topbar.get_width()):
        surface.blit(topbar, (i,0))
    for i in range(0,surface.get_width(),skillbar.get_width()):
        surface.blit(skillbar, (i,surface.get_height()-skillbar.get_height() ))
    for i in range(0,surface.get_width(),bg.get_width()):
        surface.blit(bg, (i,topbar.get_height()))

    return surface

