
# The following giant chunk is borrowed from binhex.py of Python 2.4.3.
# covox added a workaround to dump the data fork and resource fork as strings.
# Do you know what the Python one did? It didn't give the resource fork on a non-Macintosh! 
# What is the point having it it in the Python standard library? It just sits there and 
# when you need to open a BinHex file - when you actually need the bloody thing - it 
# doesn't give you half the bloody file! I mean what is happening to this country! It's 
# bloody Wilson!

# Original by Jack Jansen, CWI, August 1995.

import sys
import os
import struct
import binascii


class Error(Exception):
    pass

# States (what have we written)
[_DID_HEADER, _DID_DATA, _DID_RSRC] = range(3)

# Various constants
REASONABLY_LARGE=32768  # Minimal amount we pass the rle-coder
LINELEN=64
RUNCHAR=chr(0x90)   # run-length introducer

#
# This code is no longer byte-order dependent

#
# Workarounds for non-mac machines.
if os.name == 'mac':
    import macfs
    import MacOS
    try:
        openrf = MacOS.openrf
    except AttributeError:
        # Backward compatibility
        openrf = open

    def FInfo():
        return macfs.FInfo()

    def getfileinfo(name):
        finfo = macfs.FSSpec(name).GetFInfo()
        dir, file = os.path.split(name)
        # XXXX Get resource/data sizes
        fp = open(name, 'rb')
        fp.seek(0, 2)
        dlen = fp.tell()
        fp = openrf(name, '*rb')
        fp.seek(0, 2)
        rlen = fp.tell()
        return file, finfo, dlen, rlen

    def openrsrc(name, *mode):
        if not mode:
            mode = '*rb'
        else:
            mode = '*' + mode[0]
        return openrf(name, mode)

else:
    #
    # Glue code for non-macintosh usage
    #

    class FInfo:
        def __init__(self):
            self.Type = '????'
            self.Creator = '????'
            self.Flags = 0

    def getfileinfo(name):
        finfo = FInfo()
        # Quick check for textfile
        fp = open(name)
        data = open(name).read(256)
        for c in data:
            if not c.isspace() and (c<' ' or ord(c) > 0x7f):
                break
        else:
            finfo.Type = 'TEXT'
        fp.seek(0, 2)
        dsize = fp.tell()
        fp.close()
        dir, file = os.path.split(name)
        file = file.replace(':', '-', 1)
        return file, finfo, dsize, 0

    class openrsrc:
        def __init__(self, *args):
            pass

        def read(self, *args):
            return ''

        def write(self, *args):
            pass

        def close(self):
            pass


class _Rlecoderengine:
    """Write data to the RLE-coder in suitably large chunks"""

    def __init__(self, ofp):
        self.ofp = ofp
        self.data = ''

    def write(self, data):
        self.data = self.data + data
        if len(self.data) < REASONABLY_LARGE:
            return
        rledata = binascii.rlecode_hqx(self.data)
        self.ofp.write(rledata)
        self.data = ''

    def close(self):
        if self.data:
            rledata = binascii.rlecode_hqx(self.data)
            self.ofp.write(rledata)
        self.ofp.close()
        del self.ofp



class _Hqxdecoderengine:
    """Read data via the decoder in 4-byte chunks"""

    def __init__(self, ifp):
        self.ifp = ifp
        self.eof = 0

    def read(self, totalwtd):
        """Read at least wtd bytes (or until EOF)"""
        decdata = ''
        wtd = totalwtd
        #
        # The loop here is convoluted, since we don't really now how
        # much to decode: there may be newlines in the incoming data.
        while wtd > 0:
            if self.eof: return decdata
            wtd = ((wtd+2)//3)*4
            data = self.ifp.read(wtd)
            #
            # Next problem: there may not be a complete number of
            # bytes in what we pass to a2b. Solve by yet another
            # loop.
            #
            while 1:
                try:
                    decdatacur, self.eof = \
                            binascii.a2b_hqx(data)
                    break
                except binascii.Incomplete:
                    pass
                newdata = self.ifp.read(1)
                if not newdata:
                    raise Error, \
                          'Premature EOF on binhex file'
                data = data + newdata
            decdata = decdata + decdatacur
            wtd = totalwtd - len(decdata)
            if not decdata and not self.eof:
                raise Error, 'Premature EOF on binhex file'
        return decdata

    def close(self):
        self.ifp.close()

class _Rledecoderengine:
    """Read data via the RLE-coder"""

    def __init__(self, ifp):
        self.ifp = ifp
        self.pre_buffer = ''
        self.post_buffer = ''
        self.eof = 0

    def read(self, wtd):
        if wtd > len(self.post_buffer):
            self._fill(wtd-len(self.post_buffer))
        rv = self.post_buffer[:wtd]
        self.post_buffer = self.post_buffer[wtd:]
        return rv

    def _fill(self, wtd):
        self.pre_buffer = self.pre_buffer + self.ifp.read(wtd+4)
        if self.ifp.eof:
            self.post_buffer = self.post_buffer + \
                binascii.rledecode_hqx(self.pre_buffer)
            self.pre_buffer = ''
            return

        #
        # Obfuscated code ahead. We have to take care that we don't
        # end up with an orphaned RUNCHAR later on. So, we keep a couple
        # of bytes in the buffer, depending on what the end of
        # the buffer looks like:
        # '\220\0\220' - Keep 3 bytes: repeated \220 (escaped as \220\0)
        # '?\220' - Keep 2 bytes: repeated something-else
        # '\220\0' - Escaped \220: Keep 2 bytes.
        # '?\220?' - Complete repeat sequence: decode all
        # otherwise: keep 1 byte.
        #
        mark = len(self.pre_buffer)
        if self.pre_buffer[-3:] == RUNCHAR + '\0' + RUNCHAR:
            mark = mark - 3
        elif self.pre_buffer[-1] == RUNCHAR:
            mark = mark - 2
        elif self.pre_buffer[-2:] == RUNCHAR + '\0':
            mark = mark - 2
        elif self.pre_buffer[-2] == RUNCHAR:
            pass # Decode all
        else:
            mark = mark - 1

        self.post_buffer = self.post_buffer + \
            binascii.rledecode_hqx(self.pre_buffer[:mark])
        self.pre_buffer = self.pre_buffer[mark:]

    def close(self):
        self.ifp.close()

class HexBin:
    def __init__(self, ifp):
        if type(ifp) == type(''):
            ifp = open(ifp)
        #
        # Find initial colon.
        #
        while 1:
            ch = ifp.read(1)
            if not ch:
                raise Error, "No binhex data found"
            if ch == ':':
                break

        hqxifp = _Hqxdecoderengine(ifp)
        self.ifp = _Rledecoderengine(hqxifp)
        self.crc = 0
        self._readheader()

    def _read(self, len):
        data = self.ifp.read(len)
        self.crc = binascii.crc_hqx(data, self.crc)
        return data

    def _checkcrc(self):
        filecrc = struct.unpack('>h', self.ifp.read(2))[0] & 0xffff
        #self.crc = binascii.crc_hqx('\0\0', self.crc)
        # XXXX Is this needed??
        self.crc = self.crc & 0xffff
        if filecrc != self.crc:
            raise Error, 'CRC error, computed %x, read %x' \
                  %(self.crc, filecrc)
        self.crc = 0

    def _readheader(self):
        len = self._read(1)
        fname = self._read(ord(len))
        rest = self._read(1+4+4+2+4+4)
        self._checkcrc()

        type = rest[1:5]
        creator = rest[5:9]
        flags = struct.unpack('>h', rest[9:11])[0]
        self.dlen = struct.unpack('>l', rest[11:15])[0]
        self.rlen = struct.unpack('>l', rest[15:19])[0]

        self.FName = fname
        self.FInfo = FInfo()
        self.FInfo.Creator = creator
        self.FInfo.Type = type
        self.FInfo.Flags = flags

        self.state = _DID_HEADER

    def read(self, *n):
        if self.state != _DID_HEADER:
            raise Error, 'Read data at wrong time'
        if n:
            n = n[0]
            n = min(n, self.dlen)
        else:
            n = self.dlen
        rv = ''
        while len(rv) < n:
            rv = rv + self._read(n-len(rv))
        self.dlen = self.dlen - n
        return rv

    def close_data(self):
        if self.state != _DID_HEADER:
            raise Error, 'close_data at wrong time'
        if self.dlen:
            dummy = self._read(self.dlen)
        self._checkcrc()
        self.state = _DID_DATA

    def read_rsrc(self, *n):
        if self.state == _DID_HEADER:
            self.close_data()
        if self.state != _DID_DATA:
            raise Error, 'Read resource data at wrong time'
        if n:
            n = n[0]
            n = min(n, self.rlen)
        else:
            n = self.rlen
        self.rlen = self.rlen - n
        return self._read(n)

    def close(self):
        if self.rlen:
            dummy = self.read_rsrc(self.rlen)
        self._checkcrc()
        self.state = _DID_RSRC
        self.ifp.close()

def unpackHqx(inp):
    """unpackHqx(infilename) -> {"data":data, "rsrc":rsrc} - Decodes a BinHexed file"""
    result = {"data":"","rsrc":""}
    ifp = HexBin(inp)
    finfo = ifp.FInfo
    # XXXX Do translation on non-mac systems
    while 1:
        d = ifp.read(128000)
        if not d: break
        result["data"] += d

    ifp.close_data()

    d = ifp.read_rsrc(128000)
    if d:
        result["rsrc"]+= d
        while 1:
            d = ifp.read_rsrc(128000)
            if not d: break
            result["rsrc"]+= d

    ifp.close()
    result["map"] = createResourceMap(result["rsrc"])
    return result




# If you can read the next bit, you have terrible life issues and should see a therapist.
# This is a truly shonky method for pythonating resource forks, based entirely on the 
# (two eggs short of a spanish omlette) resource fork spec at
# http://developer.apple.com/documentation/mac/MoreToolbox/MoreToolbox-99.html

def createResourceMap(rsrc):
    """createResourceMap(rsrc) -> resourceMap - Returns a dictionary-based Resource Map"""
    dataOffset = (ord(rsrc[0])<<24) + (ord(rsrc[1])<<16) + (ord(rsrc[2])<<8) + ord(rsrc[3])
    mapOffset = (ord(rsrc[4])<<24) + (ord(rsrc[5])<<16) + (ord(rsrc[6])<<8) + ord(rsrc[7])
    dataLen = (ord(rsrc[8])<<24) + (ord(rsrc[9])<<16) + (ord(rsrc[10])<<8) + ord(rsrc[11])
    mapLen = (ord(rsrc[12])<<24) + (ord(rsrc[13])<<16) + (ord(rsrc[14])<<8) + ord(rsrc[15])

    rawData = rsrc[dataOffset:][:dataLen]  
    rawMap = rsrc[mapOffset:][:mapLen]
    
    mapTypeOffset = (ord(rawMap[24])<<8) + ord(rawMap[25])
    mapNameOffset = (ord(rawMap[26])<<8) + ord(rawMap[27])
    numberOfTypes = (ord(rawMap[28])<<8) + ord(rawMap[29])+1
    mapListsOffset = mapTypeOffset + numberOfTypes*8

    resourceMap = {}

    for j in range(mapTypeOffset+2,mapListsOffset,8): #item in a resource type list
        amount = (ord(rawMap[j+4])<<8) + ord(rawMap[j+5])+1
        offset = (ord(rawMap[j+6])<<8) + ord(rawMap[j+7])
        
        resourceMap[rawMap[j:j+4]]=[]
        for k in range(mapTypeOffset+offset,mapTypeOffset+offset+12*amount,12): #entry in the reference list for a resource type
            instanceID = (ord(rawMap[k])<<8)+ord(rawMap[k+1])
            if ((ord(rawMap[k+2])<<8) + ord(rawMap[k+3]) )==65535:
                instanceName = ""
            else:
                instanceNameLen = ord(rawMap[mapNameOffset+((ord(rawMap[k+2])<<8) + ord(rawMap[k+3]) )])
                instanceName = rawMap[mapNameOffset+((ord(rawMap[k+2])<<8) + ord(rawMap[k+3]) )+1:][:instanceNameLen]
            instanceDataOffset = (ord(rawMap[k+5])<<16) + (ord(rawMap[k+6])<<8) + ord(rawMap[k+7])
            instanceDataLen = (ord(rawData[instanceDataOffset])<<24) + (ord(rawData[instanceDataOffset+1])<<16) + (ord(rawData[instanceDataOffset+2])<<8) + ord(rawData[instanceDataOffset+3])            
            instanceData = rawData[instanceDataOffset+4:][:instanceDataLen]
            resourceMap[rawMap[j:j+4]].append(( instanceID , instanceName, instanceData ))

    return resourceMap

# having a bit of a think, the data in the Graphics.hqx resource fork might be structured thus
# 4 bytes -  start offset 
# 4 bytes -  chunksize
# 4 bytes -  ???


# Decodes an apple "clut" resource
# For some ungodly reason, 256 values for each colour channel wasn't enough for apple.
# Oh no. Red green and blue each max at 65535, plus a completely redundant 2-byte 
# palette index (for heaven's sake THE COLOURS ARE IN ASCENDING CONSECUTIVE ORDER) making it 64-bit colour
# The point? Nobody knows, as it'll be downsampled. A bloody waste of disk space.

def createColourMap(clut):
    """createColourMap(clut) -> colours - changes a "clut" table into RGB values"""
    clutLength = (ord(clut[6])<<8) + ord(clut[7])+1
    colourMap = []    
    for i in range(8,8+clutLength*8,8):
        colourMap.append(( ord(clut[i+2]), ord(clut[i+4]), ord(clut[i+6]) ))
    return colourMap    

def createShpd(shpd):
    result = []    
    for m in shpd:
        result.append((m[0],m[1],(ord(m[2][0])<<24)+(ord(m[2][1])<<16)+(ord(m[2][2])<<8)+ord(m[2][3]), 
                    (ord(m[2][4])<<24)+(ord(m[2][5])<<16)+(ord(m[2][6])<<8)+ord(m[2][7]),
                    (ord(m[2][8])<<24)+(ord(m[2][9])<<16)+(ord(m[2][10])<<8)+ord(m[2][11]) ))
    return result


